@@ -12,11 +12,21 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-use std::{convert::TryFrom, sync::Arc};
+use std::{
+    convert::{TryFrom, TryInto},
+    sync::Arc,
+};
 
 use dashmap::DashMap;
 use matrix_sdk_common::{locks::Mutex, uuid::Uuid};
-use ruma::{uint, DeviceId, MilliSecondsSinceUnixEpoch, UInt, UserId};
+use ruma::{
+    events::{
+        key::verification::VerificationMethod, AnyToDeviceEvent, AnyToDeviceEventContent,
+        ToDeviceEvent,
+    },
+    serde::Raw,
+    DeviceId, DeviceIdBox, EventId, MilliSecondsSinceUnixEpoch, RoomId, UserId,
+};
 use tracing::{info, trace, warn};
 
 use super::{
@@ -24,21 +34,20 @@ use super::{
     event_enums::{AnyEvent, AnyVerificationContent, OutgoingContent},
     requests::VerificationRequest,
     sas::Sas,
-    FlowId, Verification, VerificationResult,
+    FlowId, Verification, VerificationResult, VerificationStore,
 };
 use crate::{
     olm::PrivateCrossSigningIdentity,
     requests::OutgoingRequest,
     store::{CryptoStore, CryptoStoreError},
-    OutgoingVerificationRequest, ReadOnlyAccount, ReadOnlyDevice, RoomMessageRequest,
-    ToDeviceRequest,
+    OutgoingVerificationRequest, ReadOnlyAccount, ReadOnlyDevice, ReadOnlyUserIdentity,
+    RoomMessageRequest, ToDeviceRequest,
 };
 
 #[derive(Clone, Debug)]
 pub struct VerificationMachine {
-    account: ReadOnlyAccount,
-    private_identity: Arc<Mutex<PrivateCrossSigningIdentity>>,
-    pub(crate) store: Arc<dyn CryptoStore>,
+    pub(crate) private_identity: Arc<Mutex<PrivateCrossSigningIdentity>>,
+    pub(crate) store: VerificationStore,
     verifications: VerificationCache,
     requests: Arc<DashMap<UserId, DashMap<String, VerificationRequest>>>,
 }
@@ -50,28 +59,88 @@ impl VerificationMachine {
         store: Arc<dyn CryptoStore>,
     ) -> Self {
         Self {
-            account,
             private_identity: identity,
-            store,
+            store: VerificationStore { account, inner: store },
             verifications: VerificationCache::new(),
-            requests: DashMap::new().into(),
+            requests: Default::default(),
         }
     }
 
+    pub(crate) fn own_user_id(&self) -> &UserId {
+        self.store.account.user_id()
+    }
+
+    pub(crate) fn own_device_id(&self) -> &DeviceId {
+        self.store.account.device_id()
+    }
+
+    pub(crate) async fn request_to_device_verification(
+        &self,
+        user_id: &UserId,
+        recipient_devices: Vec<DeviceIdBox>,
+        methods: Option<Vec<VerificationMethod>>,
+    ) -> (VerificationRequest, OutgoingVerificationRequest) {
+        let flow_id = FlowId::from(Uuid::new_v4().to_string());
+
+        let verification = VerificationRequest::new(
+            self.verifications.clone(),
+            self.private_identity.lock().await.clone(),
+            self.store.clone(),
+            flow_id,
+            user_id,
+            recipient_devices,
+            methods,
+        );
+
+        self.insert_request(verification.clone());
+
+        let request = verification.request_to_device();
+
+        (verification, request.into())
+    }
+
+    pub async fn request_verification(
+        &self,
+        identity: &ReadOnlyUserIdentity,
+        room_id: &RoomId,
+        request_event_id: &EventId,
+        methods: Option<Vec<VerificationMethod>>,
+    ) -> VerificationRequest {
+        let flow_id = FlowId::InRoom(room_id.to_owned(), request_event_id.to_owned());
+
+        let request = VerificationRequest::new(
+            self.verifications.clone(),
+            self.private_identity.lock().await.clone(),
+            self.store.clone(),
+            flow_id,
+            identity.user_id(),
+            vec![],
+            methods,
+        );
+
+        self.insert_request(request.clone());
+
+        request
+    }
+
     pub async fn start_sas(
         &self,
         device: ReadOnlyDevice,
     ) -> Result<(Sas, OutgoingVerificationRequest), CryptoStoreError> {
         let identity = self.store.get_user_identity(device.user_id()).await?;
+        let own_identity =
+            self.store.get_user_identity(self.own_user_id()).await?.and_then(|i| i.into_own());
         let private_identity = self.private_identity.lock().await.clone();
 
         let (sas, content) = Sas::start(
-            self.account.clone(),
             private_identity,
             device.clone(),
             self.store.clone(),
+            own_identity,
             identity,
             None,
+            true,
+            None,
         );
 
         let request = match content {
@@ -123,6 +192,8 @@ impl VerificationMachine {
 
     #[cfg(not(target_arch = "wasm32"))]
     fn is_timestamp_valid(timestamp: &MilliSecondsSinceUnixEpoch) -> bool {
+        use ruma::{uint, UInt};
+
         // The event should be ignored if the event is older than 10 minutes
         let old_timestamp_threshold: UInt = uint!(600);
         // The event should be ignored if the event is 5 minutes or more into the
@@ -162,15 +233,40 @@ impl VerificationMachine {
         self.verifications.outgoing_requests()
     }
 
-    pub fn garbage_collect(&self) {
+    pub fn garbage_collect(&self) -> Vec<Raw<AnyToDeviceEvent>> {
+        let mut events = vec![];
+
         for user_verification in self.requests.iter() {
             user_verification.retain(|_, v| !(v.is_done() || v.is_cancelled()));
         }
         self.requests.retain(|_, v| !v.is_empty());
 
-        for request in self.verifications.garbage_collect() {
-            self.verifications.add_request(request)
+        let mut requests: Vec<OutgoingVerificationRequest> = self
+            .requests
+            .iter()
+            .flat_map(|v| {
+                let requests: Vec<OutgoingVerificationRequest> =
+                    v.value().iter().filter_map(|v| v.cancel_if_timed_out()).collect();
+                requests
+            })
+            .collect();
+
+        requests.extend(self.verifications.garbage_collect().into_iter());
+
+        for request in requests {
+            if let Ok(OutgoingContent::ToDevice(AnyToDeviceEventContent::KeyVerificationCancel(
+                content,
+            ))) = request.clone().try_into()
+            {
+                let event = ToDeviceEvent { content, sender: self.own_user_id().to_owned() };
+
+                events.push(AnyToDeviceEvent::KeyVerificationCancel(event).into());
+            }
+
+            self.verifications.add_verification_request(request)
         }
+
+        events
     }
 
     async fn mark_sas_as_done(
@@ -226,8 +322,8 @@ impl VerificationMachine {
         };
 
         let event_sent_from_us = |event: &AnyEvent<'_>, from_device: &DeviceId| {
-            if event.sender() == self.account.user_id() {
-                from_device == self.account.device_id() || event.is_room_event()
+            if event.sender() == self.store.account.user_id() {
+                from_device == self.store.account.device_id() || event.is_room_event()
             } else {
                 false
             }
@@ -247,7 +343,6 @@ impl VerificationMachine {
                             if !event_sent_from_us(&event, r.from_device()) {
                                 let request = VerificationRequest::from_request(
                                     self.verifications.clone(),
-                                    self.account.clone(),
                                     self.private_identity.lock().await.clone(),
                                     self.store.clone(),
                                     event.sender(),
@@ -284,9 +379,17 @@ impl VerificationMachine {
                         verification.receive_cancel(event.sender(), c);
                     }
 
-                    if let Some(sas) = self.get_sas(event.sender(), flow_id.as_str()) {
-                        // This won't produce an outgoing content
-                        let _ = sas.receive_any_event(event.sender(), &content);
+                    if let Some(verification) =
+                        self.get_verification(event.sender(), flow_id.as_str())
+                    {
+                        match verification {
+                            Verification::SasV1(sas) => {
+                                // This won't produce an outgoing content
+                                let _ = sas.receive_any_event(event.sender(), &content);
+                            }
+                            #[cfg(feature = "qrcode")]
+                            Verification::QrV1(qr) => qr.receive_cancel(event.sender(), c),
+                        }
                     }
                 }
                 AnyVerificationContent::Ready(c) => {
@@ -312,16 +415,22 @@ impl VerificationMachine {
                             self.store.get_device(event.sender(), c.from_device()).await?
                         {
                             let private_identity = self.private_identity.lock().await.clone();
+                            let own_identity = self
+                                .store
+                                .get_user_identity(self.own_user_id())
+                                .await?
+                                .and_then(|i| i.into_own());
                             let identity = self.store.get_user_identity(event.sender()).await?;
 
                             match Sas::from_start_event(
                                 flow_id,
                                 c,
                                 self.store.clone(),
-                                self.account.clone(),
                                 private_identity,
                                 device,
+                                own_identity,
                                 identity,
+                                None,
                                 false,
                             ) {
                                 Ok(sas) => {
@@ -369,6 +478,7 @@ impl VerificationMachine {
                         verification.receive_done(event.sender(), c);
                     }
 
+                    #[allow(clippy::single_match)]
                     match self.get_verification(event.sender(), flow_id.as_str()) {
                         Some(Verification::SasV1(sas)) => {
                             let content = sas.receive_any_event(event.sender(), &content);
@@ -377,8 +487,9 @@ impl VerificationMachine {
                                 self.mark_sas_as_done(sas, content).await?;
                             }
                         }
+                        #[cfg(feature = "qrcode")]
                         Some(Verification::QrV1(qr)) => {
-                            let (cancellation, request) = qr.receive_done(&c).await?;
+                            let (cancellation, request) = qr.receive_done(c).await?;
 
                             if let Some(c) = cancellation {
                                 self.verifications.add_request(c.into())
@@ -413,10 +524,11 @@ mod test {
     use super::{Sas, VerificationMachine};
     use crate::{
         olm::PrivateCrossSigningIdentity,
-        store::{CryptoStore, MemoryStore},
+        store::MemoryStore,
         verification::{
             event_enums::{AcceptContent, KeyContent, MacContent, OutgoingContent},
             test::wrap_any_to_device_content,
+            VerificationStore,
         },
         ReadOnlyAccount, ReadOnlyDevice,
     };
@@ -449,16 +561,19 @@ mod test {
         store.save_devices(vec![bob_device]).await;
         bob_store.save_devices(vec![alice_device.clone()]).await;
 
-        let bob_store: Arc<dyn CryptoStore> = Arc::new(bob_store);
+        let bob_store = VerificationStore { account: bob, inner: Arc::new(bob_store) };
+
         let identity = Arc::new(Mutex::new(PrivateCrossSigningIdentity::empty(alice_id())));
         let machine = VerificationMachine::new(alice, identity, Arc::new(store));
         let (bob_sas, start_content) = Sas::start(
-            bob,
             PrivateCrossSigningIdentity::empty(bob_id()),
             alice_device,
             bob_store,
             None,
             None,
+            None,
+            true,
+            None,
         );
 
         machine
