@@ -31,17 +31,18 @@ use tracing::{trace, warn};
 
 use super::{FlowId, OutgoingContent};
 use crate::{
-    identities::{ReadOnlyDevice, UserIdentities},
+    identities::{ReadOnlyDevice, ReadOnlyUserIdentities},
     utilities::encode,
     verification::event_enums::{MacContent, StartContent},
-    ReadOnlyAccount,
+    ReadOnlyAccount, ReadOnlyOwnUserIdentity,
 };
 
 #[derive(Clone, Debug)]
 pub struct SasIds {
     pub account: ReadOnlyAccount,
+    pub own_identity: Option<ReadOnlyOwnUserIdentity>,
     pub other_device: ReadOnlyDevice,
-    pub other_identity: Option<UserIdentities>,
+    pub other_identity: Option<ReadOnlyUserIdentities>,
 }
 
 /// Calculate the commitment for a accept event from the public key and the
@@ -182,7 +183,7 @@ pub fn receive_mac_event(
     flow_id: &str,
     sender: &UserId,
     content: &MacContent,
-) -> Result<(Vec<ReadOnlyDevice>, Vec<UserIdentities>), CancelCode> {
+) -> Result<(Vec<ReadOnlyDevice>, Vec<ReadOnlyUserIdentities>), CancelCode> {
     let mut verified_devices = Vec::new();
     let mut verified_identities = Vec::new();
 
@@ -304,6 +305,20 @@ pub fn get_mac_content(sas: &OlmSas, ids: &SasIds, flow_id: &FlowId) -> Outgoing
         sas.calculate_mac(key, &format!("{}{}", info, key_id)).expect("Can't calculate SAS MAC"),
     );
 
+    if let Some(own_identity) = &ids.own_identity {
+        if own_identity.is_verified() {
+            if let Some(key) = own_identity.master_key().get_first_key() {
+                let key_id = format!("{}:{}", DeviceKeyAlgorithm::Ed25519, &key);
+
+                let calculated_mac = sas
+                    .calculate_mac(key, &format!("{}{}", info, &key_id))
+                    .expect("Can't calculate SAS Master key MAC");
+
+                mac.insert(key_id, calculated_mac);
+            }
+        }
+    }
+
     // TODO Add the cross signing master key here if we trust/have it.
 
     let mut keys = mac.keys().cloned().collect::<Vec<String>>();
