@@ -25,11 +25,7 @@ use matrix_sdk_common::uuid::Uuid;
 use ruma::{
     api::client::r0::keys::upload_signatures::Request as SignatureUploadRequest,
     events::{
-        key::verification::{
-            accept::{AcceptEventContent, AcceptMethod, AcceptToDeviceEventContent},
-            cancel::CancelCode,
-            ShortAuthenticationString,
-        },
+        key::verification::{cancel::CancelCode, ShortAuthenticationString},
         AnyMessageEventContent, AnyToDeviceEventContent,
     },
     DeviceId, EventId, RoomId, UserId,
@@ -38,14 +34,15 @@ use tracing::trace;
 
 use super::{
     event_enums::{AnyVerificationContent, OutgoingContent, OwnedAcceptContent, StartContent},
-    FlowId, IdentitiesBeingVerified, VerificationResult,
+    requests::RequestHandle,
+    CancelInfo, FlowId, IdentitiesBeingVerified, VerificationResult, VerificationStore,
 };
 use crate::{
-    identities::{ReadOnlyDevice, UserIdentities},
+    identities::{ReadOnlyDevice, ReadOnlyUserIdentities},
     olm::PrivateCrossSigningIdentity,
     requests::{OutgoingVerificationRequest, RoomMessageRequest},
-    store::{CryptoStore, CryptoStoreError},
-    ReadOnlyAccount, ToDeviceRequest,
+    store::CryptoStoreError,
+    ReadOnlyAccount, ReadOnlyOwnUserIdentity, ToDeviceRequest,
 };
 
 /// Short authentication string object.
@@ -55,6 +52,8 @@ pub struct Sas {
     account: ReadOnlyAccount,
     identities_being_verified: IdentitiesBeingVerified,
     flow_id: Arc<FlowId>,
+    we_started: bool,
+    request_handle: Option<RequestHandle>,
 }
 
 impl Sas {
@@ -88,6 +87,15 @@ impl Sas {
         &self.flow_id
     }
 
+    /// Get the room id if the verification is happening inside a room.
+    pub fn room_id(&self) -> Option<&RoomId> {
+        if let FlowId::InRoom(r, _) = self.flow_id() {
+            Some(r)
+        } else {
+            None
+        }
+    }
+
     /// Does this verification flow support displaying emoji for the short
     /// authentication string.
     pub fn supports_emoji(&self) -> bool {
@@ -104,6 +112,31 @@ impl Sas {
         self.identities_being_verified.is_self_verification()
     }
 
+    /// Have we confirmed that the short auth string matches.
+    pub fn have_we_confirmed(&self) -> bool {
+        self.inner.lock().unwrap().have_we_confirmed()
+    }
+
+    /// Has the verification been accepted by both parties.
+    pub fn has_been_accepted(&self) -> bool {
+        self.inner.lock().unwrap().has_been_accepted()
+    }
+
+    /// Get info about the cancellation if the verification flow has been
+    /// cancelled.
+    pub fn cancel_info(&self) -> Option<CancelInfo> {
+        if let InnerSas::Cancelled(c) = &*self.inner.lock().unwrap() {
+            Some(c.state.as_ref().clone().into())
+        } else {
+            None
+        }
+    }
+
+    /// Did we initiate the verification flow.
+    pub fn we_started(&self) -> bool {
+        self.we_started
+    }
+
     #[cfg(test)]
     #[allow(dead_code)]
     pub(crate) fn set_creation_time(&self, time: Instant) {
@@ -112,17 +145,20 @@ impl Sas {
 
     fn start_helper(
         inner_sas: InnerSas,
-        account: ReadOnlyAccount,
         private_identity: PrivateCrossSigningIdentity,
         other_device: ReadOnlyDevice,
-        store: Arc<dyn CryptoStore>,
-        other_identity: Option<UserIdentities>,
+        store: VerificationStore,
+        other_identity: Option<ReadOnlyUserIdentities>,
+        we_started: bool,
+        request_handle: Option<RequestHandle>,
     ) -> Sas {
         let flow_id = inner_sas.verification_flow_id();
 
+        let account = store.account.clone();
+
         let identities = IdentitiesBeingVerified {
             private_identity,
-            store: store.clone(),
+            store,
             device_being_verified: other_device,
             identity_being_verified: other_identity,
         };
@@ -132,6 +168,8 @@ impl Sas {
             account,
             identities_being_verified: identities,
             flow_id,
+            we_started,
+            request_handle,
         }
     }
 
@@ -145,17 +183,21 @@ impl Sas {
     ///
     /// Returns the new `Sas` object and a `StartEventContent` that needs to be
     /// sent out through the server to the other device.
+    #[allow(clippy::too_many_arguments)]
     pub(crate) fn start(
-        account: ReadOnlyAccount,
         private_identity: PrivateCrossSigningIdentity,
         other_device: ReadOnlyDevice,
-        store: Arc<dyn CryptoStore>,
-        other_identity: Option<UserIdentities>,
+        store: VerificationStore,
+        own_identity: Option<ReadOnlyOwnUserIdentity>,
+        other_identity: Option<ReadOnlyUserIdentities>,
         transaction_id: Option<String>,
+        we_started: bool,
+        request_handle: Option<RequestHandle>,
     ) -> (Sas, OutgoingContent) {
         let (inner, content) = InnerSas::start(
-            account.clone(),
+            store.account.clone(),
             other_device.clone(),
+            own_identity,
             other_identity.clone(),
             transaction_id,
         );
@@ -163,11 +205,12 @@ impl Sas {
         (
             Self::start_helper(
                 inner,
-                account,
                 private_identity,
                 other_device,
                 store,
                 other_identity,
+                we_started,
+                request_handle,
             ),
             content,
         )
@@ -183,31 +226,36 @@ impl Sas {
     ///
     /// Returns the new `Sas` object and a `StartEventContent` that needs to be
     /// sent out through the server to the other device.
+    #[allow(clippy::too_many_arguments)]
     pub(crate) fn start_in_room(
         flow_id: EventId,
         room_id: RoomId,
-        account: ReadOnlyAccount,
         private_identity: PrivateCrossSigningIdentity,
         other_device: ReadOnlyDevice,
-        store: Arc<dyn CryptoStore>,
-        other_identity: Option<UserIdentities>,
+        store: VerificationStore,
+        own_identity: Option<ReadOnlyOwnUserIdentity>,
+        other_identity: Option<ReadOnlyUserIdentities>,
+        we_started: bool,
+        request_handle: RequestHandle,
     ) -> (Sas, OutgoingContent) {
         let (inner, content) = InnerSas::start_in_room(
             flow_id,
             room_id,
-            account.clone(),
+            store.account.clone(),
             other_device.clone(),
+            own_identity,
             other_identity.clone(),
         );
 
         (
             Self::start_helper(
                 inner,
-                account,
                 private_identity,
                 other_device,
                 store,
                 other_identity,
+                we_started,
+                Some(request_handle),
             ),
             content,
         )
@@ -227,29 +275,32 @@ impl Sas {
     pub(crate) fn from_start_event(
         flow_id: FlowId,
         content: &StartContent,
-        store: Arc<dyn CryptoStore>,
-        account: ReadOnlyAccount,
+        store: VerificationStore,
         private_identity: PrivateCrossSigningIdentity,
         other_device: ReadOnlyDevice,
-        other_identity: Option<UserIdentities>,
-        started_from_request: bool,
+        own_identity: Option<ReadOnlyOwnUserIdentity>,
+        other_identity: Option<ReadOnlyUserIdentities>,
+        request_handle: Option<RequestHandle>,
+        we_started: bool,
     ) -> Result<Sas, OutgoingContent> {
         let inner = InnerSas::from_start_event(
-            account.clone(),
+            store.account.clone(),
             other_device.clone(),
             flow_id,
             content,
+            own_identity,
             other_identity.clone(),
-            started_from_request,
+            request_handle.is_some(),
         )?;
 
         Ok(Self::start_helper(
             inner,
-            account,
             private_identity,
             other_device,
             store,
             other_identity,
+            we_started,
+            request_handle,
         ))
     }
 
@@ -271,18 +322,28 @@ impl Sas {
         &self,
         settings: AcceptSettings,
     ) -> Option<OutgoingVerificationRequest> {
-        self.inner.lock().unwrap().accept().map(|c| match settings.apply(c) {
-            OwnedAcceptContent::ToDevice(c) => {
-                let content = AnyToDeviceEventContent::KeyVerificationAccept(c);
-                self.content_to_request(content).into()
-            }
-            OwnedAcceptContent::Room(room_id, content) => RoomMessageRequest {
-                room_id,
-                txn_id: Uuid::new_v4(),
-                content: AnyMessageEventContent::KeyVerificationAccept(content),
-            }
-            .into(),
-        })
+        let mut guard = self.inner.lock().unwrap();
+        let sas: InnerSas = (*guard).clone();
+        let methods = settings.allowed_methods;
+
+        if let Some((sas, content)) = sas.accept(methods) {
+            *guard = sas;
+
+            Some(match content {
+                OwnedAcceptContent::ToDevice(c) => {
+                    let content = AnyToDeviceEventContent::KeyVerificationAccept(c);
+                    self.content_to_request(content).into()
+                }
+                OwnedAcceptContent::Room(room_id, content) => RoomMessageRequest {
+                    room_id,
+                    txn_id: Uuid::new_v4(),
+                    content: AnyMessageEventContent::KeyVerificationAccept(content),
+                }
+                .into(),
+            })
+        } else {
+            None
+        }
     }
 
     /// Confirm the Sas verification.
@@ -349,10 +410,28 @@ impl Sas {
         self.cancel_with_code(CancelCode::User)
     }
 
-    pub(crate) fn cancel_with_code(&self, code: CancelCode) -> Option<OutgoingVerificationRequest> {
+    /// Cancel the verification.
+    ///
+    /// This cancels the verification with given `CancelCode`.
+    ///
+    /// **Note**: This method should generally not be used, the [`cancel()`]
+    /// method should be preferred. The SDK will automatically cancel with the
+    /// approprate cancel code, user initiated cancellations should only cancel
+    /// with the `CancelCode::User`
+    ///
+    /// Returns None if the `Sas` object is already in a canceled state,
+    /// otherwise it returns a request that needs to be sent out.
+    ///
+    /// [`cancel()`]: #method.cancel
+    pub fn cancel_with_code(&self, code: CancelCode) -> Option<OutgoingVerificationRequest> {
         let mut guard = self.inner.lock().unwrap();
+
+        if let Some(request) = &self.request_handle {
+            request.cancel_with_code(&code)
+        }
+
         let sas: InnerSas = (*guard).clone();
-        let (sas, content) = sas.cancel(code);
+        let (sas, content) = sas.cancel(true, code);
         *guard = sas;
         content.map(|c| match c {
             OutgoingContent::Room(room_id, content) => {
@@ -436,7 +515,7 @@ impl Sas {
         self.inner.lock().unwrap().verified_devices()
     }
 
-    pub(crate) fn verified_identities(&self) -> Option<Arc<[UserIdentities]>> {
+    pub(crate) fn verified_identities(&self) -> Option<Arc<[ReadOnlyUserIdentities]>> {
         self.inner.lock().unwrap().verified_identities()
     }
 
@@ -472,23 +551,6 @@ impl AcceptSettings {
     pub fn with_allowed_methods(methods: Vec<ShortAuthenticationString>) -> Self {
         Self { allowed_methods: methods }
     }
-
-    fn apply(self, mut content: OwnedAcceptContent) -> OwnedAcceptContent {
-        match &mut content {
-            OwnedAcceptContent::ToDevice(AcceptToDeviceEventContent {
-                method: AcceptMethod::SasV1(c),
-                ..
-            })
-            | OwnedAcceptContent::Room(
-                _,
-                AcceptEventContent { method: AcceptMethod::SasV1(c), .. },
-            ) => {
-                c.short_authentication_string.retain(|sas| self.allowed_methods.contains(sas));
-                content
-            }
-            _ => content,
-        }
-    }
 }
 
 #[cfg(test)]
@@ -500,9 +562,10 @@ mod test {
     use super::Sas;
     use crate::{
         olm::PrivateCrossSigningIdentity,
-        store::{CryptoStore, MemoryStore},
-        verification::event_enums::{
-            AcceptContent, KeyContent, MacContent, OutgoingContent, StartContent,
+        store::MemoryStore,
+        verification::{
+            event_enums::{AcceptContent, KeyContent, MacContent, OutgoingContent, StartContent},
+            VerificationStore,
         },
         ReadOnlyAccount, ReadOnlyDevice,
     };
@@ -531,20 +594,23 @@ mod test {
         let bob = ReadOnlyAccount::new(&bob_id(), &bob_device_id());
         let bob_device = ReadOnlyDevice::from_account(&bob).await;
 
-        let alice_store: Arc<dyn CryptoStore> = Arc::new(MemoryStore::new());
-        let bob_store = MemoryStore::new();
+        let alice_store =
+            VerificationStore { account: alice.clone(), inner: Arc::new(MemoryStore::new()) };
 
+        let bob_store = MemoryStore::new();
         bob_store.save_devices(vec![alice_device.clone()]).await;
 
-        let bob_store: Arc<dyn CryptoStore> = Arc::new(bob_store);
+        let bob_store = VerificationStore { account: bob.clone(), inner: Arc::new(bob_store) };
 
         let (alice, content) = Sas::start(
-            alice,
             PrivateCrossSigningIdentity::empty(alice_id()),
             bob_device,
             alice_store,
             None,
             None,
+            None,
+            true,
+            None,
         );
 
         let flow_id = alice.flow_id().to_owned();
@@ -554,10 +620,11 @@ mod test {
             flow_id,
             &content,
             bob_store,
-            bob,
             PrivateCrossSigningIdentity::empty(bob_id()),
             alice_device,
             None,
+            None,
+            None,
             false,
         )
         .unwrap();
